@@ -1,11 +1,13 @@
 use crate::{
-    body::{boxed, Body, Empty, HttpBody},
+    body::{Body, HttpBody},
     response::Response,
+    util::AxumMutex,
 };
+use axum_core::{extract::Request, response::IntoResponse};
 use bytes::Bytes;
 use http::{
     header::{self, CONTENT_LENGTH},
-    HeaderValue, Request,
+    HeaderMap, HeaderValue,
 };
 use pin_project_lite::pin_project;
 use std::{
@@ -16,53 +18,78 @@ use std::{
     task::{Context, Poll},
 };
 use tower::{
-    util::{BoxCloneService, Oneshot},
+    util::{BoxCloneService, MapErrLayer, MapRequestLayer, MapResponseLayer, Oneshot},
     ServiceExt,
 };
+use tower_layer::Layer;
 use tower_service::Service;
 
 /// How routes are stored inside a [`Router`](super::Router).
 ///
 /// You normally shouldn't need to care about this type. It's used in
 /// [`Router::layer`](super::Router::layer).
-pub struct Route<B = Body, E = Infallible>(BoxCloneService<Request<B>, Response, E>);
+pub struct Route<E = Infallible>(AxumMutex<BoxCloneService<Request, Response, E>>);
 
-impl<B, E> Route<B, E> {
-    pub(super) fn new<T>(svc: T) -> Self
+impl<E> Route<E> {
+    pub(crate) fn new<T>(svc: T) -> Self
     where
-        T: Service<Request<B>, Response = Response, Error = E> + Clone + Send + 'static,
+        T: Service<Request, Error = E> + Clone + Send + 'static,
+        T::Response: IntoResponse + 'static,
         T::Future: Send + 'static,
     {
-        Self(BoxCloneService::new(svc))
+        Self(AxumMutex::new(BoxCloneService::new(
+            svc.map_response(IntoResponse::into_response),
+        )))
     }
 
     pub(crate) fn oneshot_inner(
         &mut self,
-        req: Request<B>,
-    ) -> Oneshot<BoxCloneService<Request<B>, Response, E>, Request<B>> {
-        self.0.clone().oneshot(req)
+        req: Request,
+    ) -> Oneshot<BoxCloneService<Request, Response, E>, Request> {
+        self.0.get_mut().unwrap().clone().oneshot(req)
+    }
+
+    pub(crate) fn layer<L, NewError>(self, layer: L) -> Route<NewError>
+    where
+        L: Layer<Route<E>> + Clone + Send + 'static,
+        L::Service: Service<Request> + Clone + Send + 'static,
+        <L::Service as Service<Request>>::Response: IntoResponse + 'static,
+        <L::Service as Service<Request>>::Error: Into<NewError> + 'static,
+        <L::Service as Service<Request>>::Future: Send + 'static,
+        NewError: 'static,
+    {
+        let layer = (
+            MapRequestLayer::new(|req: Request<_>| req.map(Body::new)),
+            MapErrLayer::new(Into::into),
+            MapResponseLayer::new(IntoResponse::into_response),
+            layer,
+        );
+
+        Route::new(layer.layer(self))
     }
 }
 
-impl<ReqBody, E> Clone for Route<ReqBody, E> {
+impl<E> Clone for Route<E> {
+    #[track_caller]
     fn clone(&self) -> Self {
-        Self(self.0.clone())
+        Self(AxumMutex::new(self.0.lock().unwrap().clone()))
     }
 }
 
-impl<ReqBody, E> fmt::Debug for Route<ReqBody, E> {
+impl<E> fmt::Debug for Route<E> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("Route").finish()
     }
 }
 
-impl<B, E> Service<Request<B>> for Route<B, E>
+impl<B, E> Service<Request<B>> for Route<E>
 where
-    B: HttpBody,
+    B: HttpBody<Data = bytes::Bytes> + Send + 'static,
+    B::Error: Into<axum_core::BoxError>,
 {
     type Response = Response;
     type Error = E;
-    type Future = RouteFuture<B, E>;
+    type Future = RouteFuture<E>;
 
     #[inline]
     fn poll_ready(&mut self, _cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
@@ -71,15 +98,16 @@ where
 
     #[inline]
     fn call(&mut self, req: Request<B>) -> Self::Future {
+        let req = req.map(Body::new);
         RouteFuture::from_future(self.oneshot_inner(req))
     }
 }
 
 pin_project! {
     /// Response future for [`Route`].
-    pub struct RouteFuture<B, E> {
+    pub struct RouteFuture<E> {
         #[pin]
-        kind: RouteFutureKind<B, E>,
+        kind: RouteFutureKind<E>,
         strip_body: bool,
         allow_header: Option<Bytes>,
     }
@@ -87,12 +115,12 @@ pin_project! {
 
 pin_project! {
     #[project = RouteFutureKindProj]
-    enum RouteFutureKind<B, E> {
+    enum RouteFutureKind<E> {
         Future {
             #[pin]
             future: Oneshot<
-                BoxCloneService<Request<B>, Response, E>,
-                Request<B>,
+                BoxCloneService<Request, Response, E>,
+                Request,
             >,
         },
         Response {
@@ -101,9 +129,9 @@ pin_project! {
     }
 }
 
-impl<B, E> RouteFuture<B, E> {
+impl<E> RouteFuture<E> {
     pub(crate) fn from_future(
-        future: Oneshot<BoxCloneService<Request<B>, Response, E>, Request<B>>,
+        future: Oneshot<BoxCloneService<Request, Response, E>, Request>,
     ) -> Self {
         Self {
             kind: RouteFutureKind::Future { future },
@@ -112,16 +140,6 @@ impl<B, E> RouteFuture<B, E> {
         }
     }
 
-    pub(super) fn from_response(response: Response) -> Self {
-        Self {
-            kind: RouteFutureKind::Response {
-                response: Some(response),
-            },
-            strip_body: false,
-            allow_header: None,
-        }
-    }
-
     pub(crate) fn strip_body(mut self, strip_body: bool) -> Self {
         self.strip_body = strip_body;
         self
@@ -133,17 +151,11 @@ impl<B, E> RouteFuture<B, E> {
     }
 }
 
-impl<B, E> Future for RouteFuture<B, E>
-where
-    B: HttpBody,
-{
+impl<E> Future for RouteFuture<E> {
     type Output = Result<Response, E>;
 
     #[inline]
     fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
-        #[derive(Clone, Copy)]
-        struct AlreadyPassedThroughRouteFuture;
-
         let this = self.project();
 
         let mut res = match this.kind.project() {
@@ -157,23 +169,13 @@ where
             }
         };
 
-        if res
-            .extensions()
-            .get::<AlreadyPassedThroughRouteFuture>()
-            .is_some()
-        {
-            return Poll::Ready(Ok(res));
-        } else {
-            res.extensions_mut().insert(AlreadyPassedThroughRouteFuture);
-        }
-
-        set_allow_header(&mut res, this.allow_header);
+        set_allow_header(res.headers_mut(), this.allow_header);
 
         // make sure to set content-length before removing the body
-        set_content_length(&mut res);
+        set_content_length(res.size_hint(), res.headers_mut());
 
         let res = if *this.strip_body {
-            res.map(|_| boxed(Empty::new()))
+            res.map(|_| Body::empty())
         } else {
             res
         };
@@ -182,10 +184,10 @@ where
     }
 }
 
-fn set_allow_header<B>(res: &mut Response<B>, allow_header: &mut Option<Bytes>) {
+fn set_allow_header(headers: &mut HeaderMap, allow_header: &mut Option<Bytes>) {
     match allow_header.take() {
-        Some(allow_header) if !res.headers().contains_key(header::ALLOW) => {
-            res.headers_mut().insert(
+        Some(allow_header) if !headers.contains_key(header::ALLOW) => {
+            headers.insert(
                 header::ALLOW,
                 HeaderValue::from_maybe_shared(allow_header).expect("invalid `Allow` header"),
             );
@@ -194,15 +196,12 @@ fn set_allow_header<B>(res: &mut Response<B>, allow_header: &mut Option<Bytes>)
     }
 }
 
-fn set_content_length<B>(res: &mut Response<B>)
-where
-    B: HttpBody,
-{
-    if res.headers().contains_key(CONTENT_LENGTH) {
+fn set_content_length(size_hint: http_body::SizeHint, headers: &mut HeaderMap) {
+    if headers.contains_key(CONTENT_LENGTH) {
         return;
     }
 
-    if let Some(size) = res.size_hint().exact() {
+    if let Some(size) = size_hint.exact() {
         let header_value = if size == 0 {
             #[allow(clippy::declare_interior_mutable_const)]
             const ZERO: HeaderValue = HeaderValue::from_static("0");
@@ -213,7 +212,34 @@ where
             HeaderValue::from_str(buffer.format(size)).unwrap()
         };
 
-        res.headers_mut().insert(CONTENT_LENGTH, header_value);
+        headers.insert(CONTENT_LENGTH, header_value);
+    }
+}
+
+pin_project! {
+    /// A [`RouteFuture`] that always yields a [`Response`].
+    pub struct InfallibleRouteFuture {
+        #[pin]
+        future: RouteFuture<Infallible>,
+    }
+}
+
+impl InfallibleRouteFuture {
+    pub(crate) fn new(future: RouteFuture<Infallible>) -> Self {
+        Self { future }
+    }
+}
+
+impl Future for InfallibleRouteFuture {
+    type Output = Response;
+
+    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
+        match futures_util::ready!(self.project().future.poll(cx)) {
+            Ok(response) => Poll::Ready(response),
+
+            #[allow(unreachable_patterns)]
+            Err(err) => match err {},
+        }
     }
 }
 
